from unittest.mock import patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, TextOrientation
from arkindex.process.models import ProcessMode, WorkerRun, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase


class TestBulkTranscriptions(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_run = cls.worker_version.worker_runs.filter(process__mode=ProcessMode.Workers).get()
        cls.custom_version = WorkerVersion.objects.get(worker__slug="custom")
        cls.local_worker_run = WorkerRun.objects.get(version=cls.custom_version, process__creator=cls.user)

    def test_bulk_transcriptions_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:transcription-bulk"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_bulk_transcriptions_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:transcription-bulk"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_bulk_transcriptions_not_found(self, filter_rights_mock):
        self.client.force_login(self.user)
        self.user.rights.all().delete()
        forbidden_element = self.corpus.elements.get(name="Volume 1, page 1r")

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:transcription-bulk"), {
                "worker_run_id": str(self.local_worker_run.id),
                "transcriptions": [
                    {
                        "element_id": str(forbidden_element.id),
                        "text": "lol",
                        "confidence": 0.4
                    },
                    {
                        "element_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                        "text": "lol",
                        "confidence": 0.4
                    }
                ],
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "transcriptions": [
                {"element_id": [f"Element {forbidden_element.id} was not found or cannot be written to."]},
                {"element_id": ["Element aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa was not found or cannot be written to."]},
            ]
        })

    def test_bulk_transcriptions_required_worker_run(self):
        """
        A worker run must be set
        """
        self.client.force_login(self.superuser)

        element1 = self.corpus.elements.get(name="Volume 2")
        element2 = self.corpus.elements.get(name="Volume 2, page 1r")
        self.assertFalse(element1.transcriptions.exists())
        self.assertFalse(element2.transcriptions.exists())

        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:transcription-bulk"), {
                "transcriptions": [
                    {
                        "element_id": str(element1.id),
                        "text": "Sneasel",
                        "confidence": 0.54
                    },
                    {
                        "element_id": str(element2.id),
                        "text": "Charizard",
                        "confidence": 0.85
                    },
                ],
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": ["This field is required."]
        })

    def test_bulk_transcriptions_requires_confidence(self):
        self.client.force_login(self.user)
        element = self.corpus.elements.first()

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:transcription-bulk"), {
                "worker_run_id": str(self.local_worker_run.id),
                "transcriptions": [
                    {
                        "element_id": str(element.id),
                        "text": "Sneasel"
                    },
                ],
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"transcriptions": [
            {"non_field_errors": [
                "The confidence field must be defined."
            ]}
        ]})

    def test_bulk_transcriptions_orientation(self):
        """
        A text orientation can be specified creating transcriptions in bulk
        """
        self.client.force_login(self.user)
        test_element = self.corpus.elements.get(name="Volume 2, page 1r")
        self.assertFalse(test_element.transcriptions.exists())

        with self.assertNumQueries(5):
            response = self.client.post(reverse("api:transcription-bulk"), {
                "worker_run_id": str(self.local_worker_run.id),
                "transcriptions": [
                    {
                        "element_id": str(test_element.id),
                        "text": "All Hail The Glow Cloud",
                        "orientation": TextOrientation.HorizontalRightToLeft.value,
                        "confidence": 0.78
                    },
                    {
                        "element_id": str(test_element.id),
                        "text": "The Glow Cloud does not need to converse with us.",
                        "orientation": TextOrientation.VerticalRightToLeft.value,
                        "confidence": 0.33
                    },
                ],
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        first_tr, second_tr = test_element.transcriptions.values_list("id", flat=True)
        self.assertEqual(response.json(), {
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [
                {
                    "id": str(first_tr),
                    "element_id": str(test_element.id),
                    "text": "All Hail The Glow Cloud",
                    "orientation": TextOrientation.HorizontalRightToLeft.value,
                    "confidence": 0.78,
                },
                {
                    "id": str(second_tr),
                    "element_id": str(test_element.id),
                    "text": "The Glow Cloud does not need to converse with us.",
                    "orientation": TextOrientation.VerticalRightToLeft.value,
                    "confidence": 0.33,
                },
            ]
        })

    def test_bulk_transcriptions_invalid_orientation(self):
        """
        An invalid text orientation value causes an error
        """
        self.client.force_login(self.user)
        test_element = self.corpus.elements.get(name="Volume 2, page 1r")
        self.assertFalse(test_element.transcriptions.exists())

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:transcription-bulk"), {
                "worker_run_id": str(self.local_worker_run.id),
                "transcriptions": [
                    {
                        "element_id": str(test_element.id),
                        "text": "All Hail The Glow Cloud",
                        "orientation": "khoshek",
                        "confidence": 0.78
                    },
                    {
                        "element_id": str(test_element.id),
                        "text": "The Glow Cloud does not need to converse with us.",
                        "orientation": TextOrientation.VerticalRightToLeft.value,
                        "confidence": 0.33
                    },
                ],
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "transcriptions": [
                {"orientation": ["Value is not of type TextOrientation"]},
                {}
            ]
        })

    def test_worker_run_non_local(self):
        """
        A regular user cannot create transcriptions with a WorkerRun of a non-local process
        """
        element = self.corpus.elements.get(name="Volume 2, page 1r")
        self.client.force_login(self.superuser)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:transcription-bulk"),
                data={
                    "worker_run_id": str(self.worker_run.id),
                    "transcriptions": [
                        {
                            "element_id": str(element.id),
                            "text": "Tapu Bulu",
                            "confidence": 0.42
                        },
                    ],
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Ponos task authentication is required to use a WorkerRun "
                "of a process other than the user's local process."
            ]
        })

    def test_worker_run_other_user(self):
        """
        A regular user cannot create transcriptions with a WorkerRun of someone else's local process
        """
        element = self.corpus.elements.get(name="Volume 2, page 1r")
        private_worker_run = self.user.processes.get(mode=ProcessMode.Local).worker_runs.first()
        self.client.force_login(self.superuser)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:transcription-bulk"),
                data={
                    "worker_run_id": str(private_worker_run.id),
                    "transcriptions": [
                        {
                            "element_id": str(element.id),
                            "text": "Passimian",
                            "confidence": 0.42
                        },
                    ],
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Ponos task authentication is required to use a WorkerRun "
                "of a process other than the user's local process."
            ]
        })

    def test_worker_run_other_process(self):
        """
        A Ponos task cannot create transcriptions with a WorkerRun of another process
        """
        element = self.corpus.elements.get(name="Volume 2, page 1r")
        process2 = self.worker_run.process.creator.processes.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
        )
        other_worker_run = process2.worker_runs.create(version=self.worker_run.version, parents=[])
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:transcription-bulk"),
                data={
                    "worker_run_id": str(other_worker_run.id),
                    "transcriptions": [
                        {
                            "element_id": str(element.id),
                            "text": "Passimian",
                            "confidence": 0.42
                        },
                    ],
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Only the WorkerRuns of the authenticated task's process may be used."
            ]
        })

    def test_bulk_create_local(self):
        """
        A regular user can create transcriptions with a WorkerRun of their own local process
        """
        self.client.force_login(self.user)

        element1 = self.corpus.elements.get(name="Volume 2")
        element2 = self.corpus.elements.get(name="Volume 2, page 1r")
        self.assertFalse(element1.transcriptions.exists())
        self.assertFalse(element2.transcriptions.exists())

        with self.assertNumQueries(5):
            response = self.client.post(reverse("api:transcription-bulk"), {
                "worker_run_id": str(self.local_worker_run.id),
                "transcriptions": [
                    {
                        "element_id": str(element1.id),
                        "text": "Sneasel",
                        "confidence": 0.54
                    },
                    {
                        "element_id": str(element2.id),
                        "text": "Charizard",
                        "confidence": 0.85
                    },
                    {
                        "element_id": str(element1.id),
                        "text": "Raticate",
                        "confidence": 0.12
                    },
                ],
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        first_tr, second_tr = element1.transcriptions.values_list("id", flat=True)
        third_tr = element2.transcriptions.values_list("id", flat=True)[0]
        self.assertEqual(response.json(), {
            "worker_run_id": str(self.local_worker_run.id),
            "transcriptions": [
                {
                    "id": str(first_tr),
                    "element_id": str(element1.id),
                    "text": "Sneasel",
                    "orientation": TextOrientation.HorizontalLeftToRight.value,
                    "confidence": 0.54,
                },
                {
                    "id": str(third_tr),
                    "element_id": str(element2.id),
                    "text": "Charizard",
                    "orientation": TextOrientation.HorizontalLeftToRight.value,
                    "confidence": 0.85,
                },
                {
                    "id": str(second_tr),
                    "element_id": str(element1.id),
                    "text": "Raticate",
                    "orientation": TextOrientation.HorizontalLeftToRight.value,
                    "confidence": 0.12,
                },
            ]
        })

        self.assertCountEqual(
            list(element1.transcriptions.values("text", "confidence", "worker_version_id", "worker_run_id")),
            [
                {
                    "text": "Sneasel",
                    "confidence": 0.54,
                    "worker_version_id": self.custom_version.id,
                    "worker_run_id": self.local_worker_run.id,
                },
                {
                    "text": "Raticate",
                    "confidence": 0.12,
                    "worker_version_id": self.custom_version.id,
                    "worker_run_id": self.local_worker_run.id,
                },
            ]
        )
        self.assertCountEqual(
            list(element2.transcriptions.values("text", "confidence", "worker_version_id", "worker_run_id")),
            [{
                "text": "Charizard",
                "confidence": 0.85,
                "worker_version_id": self.custom_version.id,
                "worker_run_id": self.local_worker_run.id,
            }]
        )

    def test_bulk_create_task_auth(self):
        """
        Transcriptions can be created with a WorkerRun of a non-local process
        when authenticated as a Ponos task of this process
        """
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        element = self.corpus.elements.get(name="Volume 2, page 1r")
        self.assertFalse(element.transcriptions.exists())

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-bulk"),
                data={
                    "worker_run_id": str(self.worker_run.id),
                    "transcriptions": [
                        {
                            "element_id": str(element.id),
                            "text": "Zarude",
                            "confidence": 0.12
                        },
                        {
                            "element_id": str(element.id),
                            "text": "Clefairy",
                            "confidence": 0.34
                        },
                    ],
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(
            list(element.transcriptions.values("text", "confidence", "worker_version_id", "worker_run_id")),
            [
                {
                    "text": "Zarude",
                    "confidence": 0.12,
                    "worker_version_id": self.worker_version.id,
                    "worker_run_id": self.worker_run.id,
                },
                {
                    "text": "Clefairy",
                    "confidence": 0.34,
                    "worker_version_id": self.worker_version.id,
                    "worker_run_id": self.worker_run.id,
                },
            ]
        )

    def test_worker_run_local_task_auth(self):
        """
        Transcriptions can be created with a WorkerRun of a Local process
        even when authenticated as a Ponos task from a different process
        """
        local_process = self.user.processes.get(mode=ProcessMode.Local)
        local_worker_run = local_process.worker_runs.get()
        element = self.corpus.elements.get(name="Volume 2, page 1r")

        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-bulk"),
                data={
                    "worker_run_id": str(local_worker_run.id),
                    "transcriptions": [
                        {
                            "element_id": str(element.id),
                            "text": "Chikorita",
                            "confidence": 0.42
                        },
                    ],
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(
            list(element.transcriptions.values("text", "confidence", "worker_version_id", "worker_run_id")),
            [
                {
                    "text": "Chikorita",
                    "confidence": 0.42,
                    "worker_version_id": local_worker_run.version.id,
                    "worker_run_id": local_worker_run.id,
                },
            ]
        )
