import uuid
from unittest.mock import patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Classification, ClassificationState, Corpus, Element, MLClass
from arkindex.process.models import ProcessMode, WorkerRun, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestClassifications(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.text = MLClass.objects.create(name="text", corpus=cls.corpus)
        cls.private_corpus = Corpus.objects.create(name="private")
        cls.act_type = cls.corpus.types.get(slug="act")
        cls.element = Element.objects.get(name="Volume 1, page 1v")
        cls.folder = cls.corpus.elements.get(name="Volume 1")
        cls.worker_version_1 = WorkerVersion.objects.get(worker__slug="dla")
        cls.worker_version_2 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_run = cls.worker_version_1.worker_runs.filter(process__mode=ProcessMode.Workers).get()
        cls.custom_version = WorkerVersion.objects.get(worker__slug="custom")
        cls.local_worker_run = WorkerRun.objects.get(version=cls.custom_version, process__creator=cls.user)

    def test_create_manual(self):
        """
        Creating a manual classification set auto fields correctly
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse("api:classification-create"), {
                "element": str(self.element.id),
                "ml_class": str(self.text.id),
                "worker_version": "",
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        classification = self.element.classifications.get()

        self.assertEqual(classification.worker_version, None)
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Validated)
        self.assertEqual(classification.confidence, 1)
        self.assertTrue(classification.high_confidence)
        self.assertDictEqual(response.json(), {
            "id": str(classification.id),
            "element": str(self.element.id),
            "ml_class": str(self.text.id),
            "worker_run_id": None,
            "state": ClassificationState.Validated.value,
            "confidence": 1.0,
            "high_confidence": True,
        })

    def test_create_null_worker_run(self):
        """
        A manual classification may be created specifying the worker run ID as null
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:classification-create"),
                data={
                    "element": str(self.element.id),
                    "ml_class": str(self.text.id),
                    "worker_run_id": None,
                },
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_manual_exists(self):
        """
        If a manual classification exists, creation must respond a 400_BAD_REQUEST
        with an explicit message
        """
        self.client.force_login(self.user)
        request = (
            reverse("api:classification-create"),
            {"element": str(self.element.id), "ml_class": str(self.text.id)}
        )
        with self.assertNumQueries(6):
            response = self.client.post(*request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        with self.assertNumQueries(5):
            response = self.client.post(*request)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "non_field_errors": ["A manual classification already exists for this element and this class."]
        })

    def test_create_worker_run_exists(self):
        """
        If a classification from the same worker run already exists,
        creation must respond a 400_BAD_REQUEST with an explicit message
        """
        self.client.force_login(self.user)
        request = (
            reverse("api:classification-create"),
            {
                "element": str(self.element.id),
                "ml_class": str(self.text.id),
                "worker_run_id": str(self.local_worker_run.id),
                "confidence": "0.42",
                "high_confidence": False,
            }
        )
        self.assertFalse(self.element.classifications.filter(ml_class=self.text, worker_run=self.local_worker_run).exists())
        response = self.client.post(*request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertTrue(self.element.classifications.filter(ml_class=self.text, worker_run=self.local_worker_run).exists())

        with self.assertNumQueries(6):
            response = self.client.post(*request)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "non_field_errors": ["A classification from this worker run already exists for this element and this class."]
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_create_writable_corpus(self, filter_rights_mock):
        """
        Require the element and ML class to be in a writable corpus
        """
        self.client.force_login(self.user)
        ml_class = self.private_corpus.ml_classes.create(name="Heatmor")
        self.element.corpus = self.private_corpus
        self.element.save()

        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:classification-create"), {
                "element": str(self.element.id),
                "ml_class": str(ml_class.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "element": [f'Invalid pk "{self.element.id}" - object does not exist.'],
            "ml_class": [f'Invalid pk "{ml_class.id}" - object does not exist.'],
        })

    def test_create_same_corpus(self):
        """
        Require the ML class to be in the same corpus as the element
        """
        self.client.force_login(self.user)
        self.private_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        ml_class = self.private_corpus.ml_classes.create(name="Heatmor")

        with self.assertNumQueries(5):
            response = self.client.post(reverse("api:classification-create"), {
                "element": str(self.element.id),
                "ml_class": str(ml_class.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "non_field_errors": ["Element and ML class are not in the same corpus"]
        })

    def test_create_ignored_params(self):
        """
        Ensure confidence and high_confidence attributes cannot be changed on manual classifications
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(reverse("api:classification-create"), {
                "element": str(self.element.id),
                "ml_class": str(self.text.id),
                "confidence": 0.5,
                "high_confidence": False,
                "state": "rejected",
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        classification = self.element.classifications.get()
        self.assertDictEqual(response.json(), {
            "id": str(classification.id),
            "element": str(self.element.id),
            "ml_class": str(self.text.id),
            "worker_run_id": None,
            "state": ClassificationState.Validated.value,
            "confidence": 1,
            "high_confidence": True,
        })

    def test_create_without_permission(self):
        """
        Without permission, should not raise a 403 Forbidden: indicates that the resource exists
        Indicates that page PK does not exist instead
        """
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:classification-create"), {
                "element": str(self.element.id),
                "ml_class": str(self.text.id),
            })
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_worker_run_local(self):
        """
        A regular user can create a classification with a WorkerRun of their own local process
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(reverse("api:classification-create"), {
                "element": str(self.element.id),
                "ml_class": str(self.text.id),
                "worker_run_id": str(self.local_worker_run.id),
                "confidence": 0.42,
                "high_confidence": False,
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        classification = self.element.classifications.get()
        self.assertEqual(classification.worker_run, self.local_worker_run)
        self.assertEqual(classification.worker_version, self.custom_version)
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Pending)
        self.assertEqual(classification.confidence, 0.42)
        self.assertFalse(classification.high_confidence)

        self.assertDictEqual(response.json(), {
            "id": str(classification.id),
            "element": str(self.element.id),
            "ml_class": str(self.text.id),
            "worker_run_id": str(self.local_worker_run.id),
            "state": ClassificationState.Pending.value,
            "confidence": 0.42,
            "high_confidence": False,
        })

    def test_create_worker_run_local_task_auth(self):
        """
        A classification can be created with a WorkerRun of a Local process
        even when authenticated as a Ponos task from a different process
        """
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:classification-create"),
                {
                    "element": str(self.element.id),
                    "ml_class": str(self.text.id),
                    "worker_run_id": str(self.local_worker_run.id),
                    "confidence": 0.42,
                    "high_confidence": False,
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        classification = self.element.classifications.get()
        self.assertEqual(classification.worker_run, self.local_worker_run)
        self.assertEqual(classification.worker_version, self.custom_version)
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Pending)
        self.assertEqual(classification.confidence, 0.42)
        self.assertFalse(classification.high_confidence)

        self.assertDictEqual(response.json(), {
            "id": str(classification.id),
            "element": str(self.element.id),
            "ml_class": str(self.text.id),
            "worker_run_id": str(self.local_worker_run.id),
            "state": ClassificationState.Pending.value,
            "confidence": 0.42,
            "high_confidence": False,
        })

    def test_create_worker_run_task(self):
        """
        A classification can be created with a WorkerRun of a non-local process
        when authenticated as a Ponos task of this process
        """
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:classification-create"),
                {
                    "element": str(self.element.id),
                    "ml_class": str(self.text.id),
                    "worker_run_id": str(self.worker_run.id),
                    "confidence": 0.42,
                    "high_confidence": False,
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        classification = self.element.classifications.get()
        self.assertEqual(classification.worker_run, self.worker_run)
        self.assertEqual(classification.worker_version, self.worker_version_1)
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Pending)
        self.assertEqual(classification.confidence, 0.42)
        self.assertFalse(classification.high_confidence)

        self.assertDictEqual(response.json(), {
            "id": str(classification.id),
            "element": str(self.element.id),
            "ml_class": str(self.text.id),
            "worker_run_id": str(self.worker_run.id),
            "state": ClassificationState.Pending.value,
            "confidence": 0.42,
            "high_confidence": False,
        })

    def test_create_worker_run_non_local(self):
        """
        A regular user cannot create a classification with a WorkerRun of a non-local process
        """
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            response = self.client.post(reverse("api:classification-create"), {
                "element": str(self.element.id),
                "ml_class": str(self.text.id),
                "worker_run_id": str(self.worker_run.id),
                "confidence": 0.42,
                "high_confidence": False,
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "worker_run_id": ["Ponos task authentication is required to use a WorkerRun of a process other than the user's local process."],
        })

    def test_create_worker_run_other_user(self):
        """
        A regular user cannot create a classification with a WorkerRun of someone else's local process
        """
        worker_run = self.user.processes.get(mode=ProcessMode.Local).worker_runs.first()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            response = self.client.post(reverse("api:classification-create"), {
                "element": str(self.element.id),
                "ml_class": str(self.text.id),
                "worker_run_id": str(worker_run.id),
                "confidence": 0.42,
                "high_confidence": False,
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "worker_run_id": ["Ponos task authentication is required to use a WorkerRun of a process other than the user's local process."],
        })

    def test_create_worker_run_other_process(self):
        """
        A Ponos task cannot create a classification with a WorkerRun of another process
        """
        # Create a process on the same user, to check that being the process creator has no effect
        process2 = self.worker_run.process.creator.processes.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
        )
        other_worker_run = process2.worker_runs.create(version=self.worker_run.version, parents=[])
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:classification-create"),
                {
                    "element": str(self.element.id),
                    "ml_class": str(self.text.id),
                    "worker_run_id": str(other_worker_run.id),
                    "confidence": 0.42,
                    "high_confidence": False,
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "worker_run_id": ["Only the WorkerRuns of the authenticated task's process may be used."],
        })

    def test_create_worker_run_not_found(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(reverse("api:classification-create"), {
                "element": str(self.element.id),
                "ml_class": str(self.text.id),
                "worker_run_id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "worker_run_id": ['Invalid pk "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa" - object does not exist.'],
        })

    def test_create_zero_confidence(self):
        """
        Ensure CreateClassification accepts a confidence of 0
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(reverse("api:classification-create"), {
                "element": str(self.element.id),
                "ml_class": str(self.text.id),
                "worker_run_id": str(self.local_worker_run.id),
                "confidence": 0,
                "high_confidence": False,
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        classification = self.element.classifications.get()
        self.assertEqual(classification.worker_version, self.custom_version)
        self.assertEqual(classification.worker_run_id, self.local_worker_run.id)
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Pending)
        self.assertEqual(classification.confidence, 0)
        self.assertFalse(classification.high_confidence)

    def test_create_manual_and_worker_run(self):
        """
        CreateClassification should allow creating the same ML class from a worker run
        and no worker run on the same element.
        """
        self.client.force_login(self.user)

        self.element.classifications.create(
            worker_version=self.worker_run.version,
            worker_run=self.worker_run,
            ml_class=self.text,
            confidence=0.5,
            high_confidence=False,
        )

        self.assertEqual(self.element.classifications.count(), 1)

        # Create a manual classification with the same class
        with self.assertNumQueries(6):
            response = self.client.post(reverse("api:classification-create"), {
                "element": str(self.element.id),
                "ml_class": str(self.text.id),
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(self.element.classifications.count(), 2)

        self.assertCountEqual(
            list(self.element.classifications.values("worker_version_id", "worker_run_id")),
            [
                {"worker_version_id": None, "worker_run_id": None},
                {"worker_version_id": self.worker_run.version.id, "worker_run_id": self.worker_run.id},
            ]
        )

    def test_validate(self):
        classification = self.element.classifications.create(
            worker_version=self.worker_version_1,
            worker_run=self.worker_run,
            ml_class=self.text,
            confidence=.1
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.put(reverse("api:classification-validate", kwargs={"pk": classification.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(classification.id),
            "worker_run": {
                "id": str(self.worker_run.id),
                "summary": "Worker Document layout analyser @ version 1"
            },
            "ml_class": {
                "id": str(classification.ml_class.id),
                "name": classification.ml_class.name
            },
            "state": ClassificationState.Validated.value,
            "confidence": classification.confidence,
            "high_confidence": False
        })

        # Ensure moderator has been set
        classification.refresh_from_db()
        self.assertEqual(classification.moderator, self.user)

    def test_validate_requires_login(self):
        classification = self.element.classifications.create(
            ml_class=self.text,
            confidence=.5,
        )
        with self.assertNumQueries(0):
            response = self.client.put(reverse("api:classification-validate", kwargs={"pk": classification.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_validate_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        classification = self.element.classifications.create(
            ml_class=self.text,
            confidence=.5,
        )
        with self.assertNumQueries(2):
            response = self.client.put(reverse("api:classification-validate", kwargs={"pk": classification.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_reject_worker_run(self):
        self.client.force_login(self.user)
        classification = self.element.classifications.create(
            worker_version=self.worker_version_1,
            worker_run=self.worker_run,
            ml_class=self.text,
            confidence=.1,
        )

        with self.assertNumQueries(6):
            response = self.client.put(reverse("api:classification-reject", kwargs={"pk": classification.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(classification.id),
            "worker_run": {
                "id": str(self.worker_run.id),
                "summary": "Worker Document layout analyser @ version 1"
            },
            "ml_class": {
                "id": str(classification.ml_class.id),
                "name": classification.ml_class.name
            },
            "state": ClassificationState.Rejected.value,
            "confidence": classification.confidence,
            "high_confidence": False
        })

        # Ensure moderator has been set
        classification.refresh_from_db()
        self.assertEqual(classification.moderator, self.user)

    def test_reject_manual_delete(self):
        """
        A classifications with no worker version should be deleted when rejected
        """
        self.client.force_login(self.user)
        classification = self.element.classifications.create(ml_class=self.text, confidence=.42)

        with self.assertNumQueries(4):
            response = self.client.put(reverse("api:classification-reject", kwargs={"pk": classification.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(Classification.DoesNotExist):
            classification.refresh_from_db()

    def test_reject_requires_login(self):
        classification = self.element.classifications.create(ml_class=self.text, confidence=.42)
        with self.assertNumQueries(0):
            response = self.client.put(reverse("api:classification-reject", kwargs={"pk": classification.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_reject_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        classification = self.element.classifications.create(ml_class=self.text, confidence=.42)
        with self.assertNumQueries(2):
            response = self.client.put(reverse("api:classification-reject", kwargs={"pk": classification.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_repeatable_moderation(self):
        """
        A rejected classification can be re-validated
        """
        self.client.force_login(self.user)
        classification = self.element.classifications.create(
            ml_class=self.text,
            confidence=.5,
            moderator=self.user,
            state=ClassificationState.Validated,
            worker_version=self.worker_version_1,
            worker_run=self.worker_run
        )

        # First try to reject
        with self.assertNumQueries(6):
            response = self.client.put(reverse("api:classification-reject", kwargs={"pk": classification.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(classification.id),
            "ml_class": {
                "id": str(classification.ml_class.id),
                "name": classification.ml_class.name
            },
            "state": ClassificationState.Rejected.value,
            "confidence": classification.confidence,
            "high_confidence": False,
            "worker_run": {
                "id": str(self.worker_run.id),
                "summary": "Worker Document layout analyser @ version 1"
            },
        })

        # Then try to validate
        with self.assertNumQueries(6):
            response = self.client.put(reverse("api:classification-validate", kwargs={"pk": classification.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(classification.id),
            "ml_class": {
                "id": str(classification.ml_class.id),
                "name": classification.ml_class.name
            },
            "state": ClassificationState.Validated.value,
            "confidence": classification.confidence,
            "high_confidence": False,
            "worker_run": {
                "id": str(self.worker_run.id),
                "summary": "Worker Document layout analyser @ version 1"
            },
        })

    def test_create_selection_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:classification-selection"),
                data={"corpus_id": self.corpus.id, "ml_class": self.text.id, "mode": "create"}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={"selection": False})
    def test_create_selection_no_selection(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:classification-selection"),
                data={"corpus_id": self.corpus.id, "ml_class": self.text.id, "mode": "create"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["Selection is not available on this instance."]})

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_create_selection_private_corpus(self, has_access_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:classification-selection"),
                data={"corpus_id": self.private_corpus.id, "ml_class": self.text.id, "mode": "create"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_create_selection_writable_corpus(self, has_access_mock):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)

        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:classification-selection"),
                data={"corpus_id": self.corpus.id, "ml_class": self.text.id, "mode": "create"}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {
            "detail": "You do not have contributor access to this corpus.",
        })

    def test_create_selection_non_existing_corpus(self):
        random_uuid = uuid.uuid4()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:classification-selection"),
                data={"corpus_id": random_uuid, "ml_class": self.text.id, "mode": "create"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "corpus_id": [f'Invalid pk "{random_uuid}" - object does not exist.'],
        })

    def test_create_selection_wrong_corpus(self):
        line = MLClass.objects.create(name="line", corpus=self.private_corpus)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:classification-selection"),
                data={"corpus_id": self.corpus.id, "ml_class": line.id, "mode": "create"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "non_field_errors": [f"Ml class {line} does not belong to the corpus {self.corpus}"]
        })

    def test_create_selection_wrong_ml_class(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:classification-selection"),
                data={"corpus_id": self.corpus.id, "ml_class": "ffffffff-ffff-ffff-ffff-ffffffffffff", "mode": "create"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "ml_class": ["Ml class ffffffff-ffff-ffff-ffff-ffffffffffff not found"]
        })

    def test_create_selection_wrong_mode(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:classification-selection"),
                data={"corpus_id": self.corpus.id, "ml_class": self.text.id, "mode": "test"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_selection_create(self):
        act_x = Element.objects.create(
            type=self.act_type,
            corpus=self.private_corpus,
            name="Act xxx",
        )
        act_y = Element.objects.create(
            type=self.act_type,
            corpus=self.private_corpus,
            name="Act yyy",
        )

        self.assertFalse(self.folder.classifications.exists())
        self.assertFalse(self.element.classifications.exists())

        self.client.force_login(self.superuser)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:elements-selection"),
                data={"ids": [self.folder.id, self.element.id, act_x.id, act_y.id]},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:elements-selection"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        results = response.json()["results"]
        self.assertEqual(len(results), 4)

        self.assertCountEqual(
            [element["id"] for element in results],
            [str(self.element.id), str(self.folder.id), str(act_x.id), str(act_y.id)],
        )

        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:classification-selection"),
                data={"corpus_id": self.corpus.id, "ml_class": self.text.id, "mode": "create"}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertTrue(self.folder.classifications.exists())
        classification = self.folder.classifications.get()
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Validated)
        self.assertIsNone(classification.worker_version)
        self.assertEqual(classification.confidence, 1)

        self.assertTrue(self.element.classifications.exists())
        classification = self.element.classifications.get()
        self.assertEqual(classification.ml_class, self.text)
        self.assertEqual(classification.state, ClassificationState.Validated)
        self.assertIsNone(classification.worker_version)
        self.assertEqual(classification.confidence, 1)

    def test_create_selection_validate(self):
        line = MLClass.objects.create(name="line", corpus=self.private_corpus)
        act_x = Element.objects.create(
            type=self.act_type,
            corpus=self.private_corpus,
            name="Act xxx",
        )
        Classification.objects.create(
            element=act_x,
            worker_version=self.worker_version_1,
            state=ClassificationState.Pending,
            high_confidence=True,
            confidence=1,
            ml_class=line
        )

        for e in [self.folder, self.element]:
            Classification.objects.create(
                element=e,
                worker_version=self.worker_version_1,
                state=ClassificationState.Pending,
                high_confidence=True,
                confidence=1,
                ml_class=self.text
            )
            Classification.objects.create(
                element=e,
                worker_version=self.worker_version_2,
                state=ClassificationState.Pending,
                high_confidence=False,
                confidence=0.4,
                ml_class=self.text
            )

        self.client.force_login(self.superuser)
        self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.folder.id, self.element.id, act_x.id]},
        )

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:classification-selection"),
                data={"corpus_id": self.corpus.id, "mode": "validate"}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        for e in [self.folder, self.element]:
            classification = e.classifications.get(state=ClassificationState.Validated)
            self.assertTrue(classification.high_confidence)
            self.assertEqual(classification.worker_version, self.worker_version_1)

            classification = e.classifications.get(state=ClassificationState.Pending)
            self.assertFalse(classification.high_confidence)
            self.assertEqual(classification.worker_version, self.worker_version_2)

        classification = act_x.classifications.get()
        self.assertEqual(classification.state, ClassificationState.Pending)

    def test_create_selection_create_requires_ml_class(self):
        self.client.force_login(self.superuser)
        self.client.post(
            reverse("api:elements-selection"),
            data={"ids": [self.folder.id, self.element.id]},
        )

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:classification-selection"),
                data={"corpus_id": self.corpus.id, "mode": "create"}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"ml_class": ["ML class is required when adding a classification"]})
