from unittest.mock import call, patch
from uuid import uuid4

from django.test import override_settings
from django.urls import reverse
from django.utils import timezone
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.models import Farm, State
from arkindex.process.models import (
    ActivityState,
    ArkindexFeature,
    FeatureUsage,
    ProcessDatasetSet,
    ProcessMode,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Dataset, Model, ModelVersionState


class TestProcessStart(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.dataset1, cls.dataset2 = Dataset.objects.filter(corpus=cls.corpus).order_by("name")

        other_corpus = Corpus.objects.create(name="Corpus 2")
        cls.other_dataset = other_corpus.datasets.create(
            name="Dead sea scrolls",
            description="Human instrumentality manual",
            creator=cls.user,
        )
        cls.other_dataset.sets.create(name="test")

        cls.recognizer = WorkerVersion.objects.get(worker__slug="reco")
        cls.dla = WorkerVersion.objects.get(worker__slug="dla")
        cls.version_gpu = WorkerVersion.objects.get(worker__slug="worker-gpu")
        cls.version_with_model = WorkerVersion.objects.get(worker__slug="generic")
        cls.model = Model.objects.create(name="My model")
        cls.model_version = cls.model.versions.create(
            state=ModelVersionState.Available,
            size=8,
            hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            archive_hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
        )

        cls.workers_process = cls.corpus.processes.create(creator=cls.user, mode=ProcessMode.Workers)
        cls.dataset_process = cls.corpus.processes.create(creator=cls.user, mode=ProcessMode.Dataset)

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_invalid_process_id(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": "12341234-1234-1234-1234-123412341234"})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(response.json(), {"detail": "No Process matches the given query."})

    def test_wrong_process_mode(self):
        self.client.force_login(self.user)

        for mode in set(ProcessMode) - {ProcessMode.Workers, ProcessMode.Dataset, ProcessMode.Local}:
            with self.subTest(mode=mode):
                self.workers_process.mode = mode
                self.workers_process.save()

                with self.assertNumQueries(5):
                    response = self.client.post(
                        reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(
                    response.json(),
                    {"__all__": ["Only a Process with Workers or Dataset mode and not already launched can be started later on"]}
                )

    def test_local_mode(self):
        self.client.force_login(self.user)
        local_process = self.user.processes.filter(mode=ProcessMode.Local).get()

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(local_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_process_already_started(self):
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.workers_process.run()
        self.assertTrue(self.workers_process.tasks.exists())
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            response.json(),
            {"__all__": ["Only a Process with Workers or Dataset mode and not already launched can be started later on"]}
        )

    @override_settings(PUBLIC_HOSTNAME="https://darkindex.lol")
    def test_without_required_model(self):
        self.workers_process.worker_runs.create(version=self.version_with_model)

        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            response.json(),
            {"model_version": [f"The following workers require a model version and none was set: {[self.version_with_model.worker.name]}"]}
        )

    @override_settings(PUBLIC_HOSTNAME="https://arkindex.localhost")
    @patch("arkindex.project.triggers.process_tasks.initialize_activity.delay")
    def test_with_required_model(self, activities_delay_mock):
        self.workers_process.worker_runs.create(version=self.version_with_model, model_version=self.model_version)
        self.assertFalse(self.workers_process.tasks.exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(18):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["id"], str(self.workers_process.id))

        self.assertEqual(activities_delay_mock.call_count, 1)
        self.assertEqual(activities_delay_mock.call_args, call(
            self.workers_process,
            user_id=self.user.id,
            description="Initializing worker activities for process Generic worker with a Model",
        ))

    def test_empty(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            response.json(),
            {"__all__": ["The process must have worker runs to be started."]},
        )

    def test_unavailable_worker_version(self):
        self.workers_process.worker_runs.create(version=self.recognizer)
        self.recognizer.state = WorkerVersionState.Error
        self.recognizer.save()
        self.assertFalse(self.workers_process.tasks.exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {"version": ["This process contains one or more unavailable worker versions and cannot be started."]},
        )

    def test_unavailable_model_version(self):
        self.workers_process.worker_runs.create(version=self.recognizer, model_version=self.model_version)
        self.model_version.state = ModelVersionState.Error
        self.model_version.save()
        self.assertFalse(self.workers_process.tasks.exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {"model_version": ["This process contains one or more unavailable model versions and cannot be started."]},
        )

    def test_archived_models(self):
        self.workers_process.worker_runs.create(version=self.recognizer, model_version=self.model_version)
        self.model.archived = timezone.now()
        self.model.save()
        self.assertFalse(self.workers_process.tasks.exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            ["Some model versions are on archived models and cannot be executed."],
        )

    def test_required_fields_no_config(self):
        # Both workers now have a required field without a default value
        self.dla.configuration["user_configuration"] = {
            "some_field": {
                "title": "pls configure me",
                "type": "bool",
                "required": True,
            }
        }
        self.dla.save()
        self.recognizer.configuration["user_configuration"] = {
            "some_field": {
                "title": "pls configure me",
                "type": "bool",
                "required": True,
            }
        }
        self.recognizer.save()

        # One worker version is used without any configuration
        self.workers_process.worker_runs.create(
            version=self.recognizer,
            configuration=None,
            model_version=None,
        )
        # The other version is used with a configuration missing the required field
        self.workers_process.worker_runs.create(
            version=self.dla,
            configuration=self.dla.worker.configurations.create(
                name="oh no",
                configuration={
                    "not_some_field": "oops",
                },
            ),
            model_version=None,
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "worker_configuration": [
                "The following workers have required configuration fields that have not been set: "
                "Document layout analyser, Recognizer",
            ],
        })

    @patch("arkindex.project.triggers.process_tasks.initialize_activity.delay")
    def test_workers(self, activities_delay_mock):
        """
        A user can start a process with no parameters.
        Default chunks, thumbnails and farm are used. Cache is disabled, and worker activities are enabled.
        """
        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        init_run = self.workers_process.worker_runs.create(version=init_elements_version)
        run = self.workers_process.worker_runs.create(version=self.recognizer, parents=[init_run.id])
        self.assertFalse(self.workers_process.tasks.exists())

        self.client.force_login(self.user)

        with self.assertNumQueries(15):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.json()["id"], str(self.workers_process.id))
        self.workers_process.refresh_from_db()
        # A default process name is set
        self.assertEqual(self.workers_process.name, "Recognizer")
        self.assertEqual(self.workers_process.state, State.Unscheduled)
        self.assertEqual(self.workers_process.farm, None)
        self.assertEqual(self.workers_process.activity_state, ActivityState.Pending)
        self.assertEqual(self.workers_process.tasks.count(), 2)
        task1, task2 = self.workers_process.tasks.order_by("slug")
        self.assertEqual(task1.slug, "initialisation")
        self.assertEqual(task2.slug, f"reco_{str(run.id)[:6]}")
        self.assertEqual(task1.command, init_elements_version.docker_command)
        self.assertEqual(activities_delay_mock.call_count, 1)
        self.assertEqual(activities_delay_mock.call_args, call(
            self.workers_process,
            user_id=self.user.id,
            description="Initializing worker activities for process Recognizer",
        ))

    def test_inconsistent_gpu_usages(self):
        # The version's gpu_usage is Disabled, so the run's use_gpu is set to False
        self.workers_process.worker_runs.create(version=self.recognizer)
        self.recognizer.gpu_usage = FeatureUsage.Required
        self.recognizer.save()
        self.dla.gpu_usage = FeatureUsage.Required
        self.dla.save()
        # The version's gpu_usage is Required, so the run's use_gpu is set to True
        self.workers_process.worker_runs.create(version=self.dla)
        self.dla.gpu_usage = FeatureUsage.Disabled
        self.dla.save()
        self.assertFalse(self.workers_process.tasks.exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            response.json(),
            [
                "Some worker runs do not require a GPU, but their worker version does.",
                "Some worker runs require a GPU, but their worker version does not support GPU usage."
            ],
        )

    def test_dataset_requires_datasets(self):
        self.dataset_process.worker_runs.create(version=self.recognizer)
        self.assertFalse(self.dataset_process.tasks.exists())

        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.dataset_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "non_field_errors": ["A dataset process cannot be started if it does not have any associated dataset sets."]
        })

    def test_dataset_requires_dataset_in_same_corpus(self):
        test_set = self.other_dataset.sets.get(name="test")
        ProcessDatasetSet.objects.create(process=self.dataset_process, set=test_set)
        self.dataset_process.worker_runs.create(version=self.recognizer)
        self.assertFalse(self.dataset_process.tasks.exists())

        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.dataset_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "non_field_errors": ["At least one of the process sets must be from the same corpus as the process."]
        })

    def test_dataset_unsupported_parameters(self):
        self.dataset_process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Dataset)
        test_set_1 = self.dataset1.sets.get(name="test")
        test_set_2 = self.dataset2.sets.get(name="test")
        ProcessDatasetSet.objects.create(process=self.dataset_process, set=test_set_1)
        ProcessDatasetSet.objects.create(process=self.dataset_process, set=test_set_2)
        self.dataset_process.worker_runs.create(version=self.recognizer)

        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.dataset_process.id)}),
                {
                    "use_cache": True,
                    # This should be ignored
                    "worker_activity": True,
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "use_cache": ["Caching is not supported on Dataset processes."],
        })

    def test_dataset(self):
        self.dataset_process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Dataset)
        test_set_1 = self.dataset1.sets.get(name="test")
        test_set_2 = self.other_dataset.sets.get(name="test")
        ProcessDatasetSet.objects.create(process=self.dataset_process, set=test_set_1)
        ProcessDatasetSet.objects.create(process=self.dataset_process, set=test_set_2)
        run = self.dataset_process.worker_runs.create(version=self.recognizer)
        self.assertFalse(self.dataset_process.tasks.exists())

        self.client.force_login(self.user)

        with self.assertNumQueries(12):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.dataset_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.json()["id"], str(self.dataset_process.id))
        self.dataset_process.refresh_from_db()
        # No default process name is set for non-workers mode processes
        self.assertEqual(self.dataset_process.name, None)
        self.assertEqual(self.dataset_process.state, State.Unscheduled)
        # Ensure default parameters are used
        self.assertEqual(self.dataset_process.farm, None)
        self.assertEqual(self.dataset_process.tasks.count(), 1)
        task = self.dataset_process.tasks.get()
        self.assertEqual(task.slug, run.task_slug)
        self.assertQuerySetEqual(self.dataset_process.sets.order_by("dataset__name"), [
            test_set_2, test_set_1
        ])

    @patch("arkindex.project.triggers.process_tasks.initialize_activity.delay")
    def test_from_docker_image(self, activities_delay_mock):
        """
        Start a process with an available WorkerVersion with a docker_image_iid
        """
        self.recognizer.docker_image_iid = "hello-world"
        self.recognizer.save()
        self.assertEqual(self.recognizer.state, WorkerVersionState.Available)

        run = self.workers_process.worker_runs.create(version=self.recognizer)
        self.assertFalse(self.workers_process.tasks.exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(18):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.json()["id"], str(self.workers_process.id))
        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.state, State.Unscheduled)
        task1, task2 = self.workers_process.tasks.order_by("slug")
        self.assertEqual(task1.slug, "initialisation")
        self.assertEqual(task2.slug, run.task_slug)

        self.assertEqual(activities_delay_mock.call_count, 1)
        self.assertEqual(activities_delay_mock.call_args, call(
            self.workers_process,
            user_id=self.user.id,
            description="Initializing worker activities for process Recognizer",
        ))

    @patch("arkindex.project.triggers.process_tasks.initialize_activity.delay")
    def test_select_farm_id(self, activities_delay_mock):
        """
        A user can specify a ponos farm to use for a process
        """
        self.workers_process.worker_runs.create(version=self.recognizer)
        farm = Farm.objects.get(name="Wheat farm")

        self.client.force_login(self.user)
        with self.assertNumQueries(19):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)}),
                {"farm": str(farm.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.state, State.Unscheduled)
        self.assertEqual(self.workers_process.farm_id, farm.id)

        self.assertEqual(activities_delay_mock.call_count, 1)
        self.assertEqual(activities_delay_mock.call_args, call(
            self.workers_process,
            user_id=self.user.id,
            description="Initializing worker activities for process Recognizer",
        ))

    @patch("arkindex.project.triggers.process_tasks.initialize_activity.delay")
    @patch("arkindex.process.serializers.imports.get_default_farm")
    def test_default_farm(self, get_default_farm_mock, activities_delay_mock):
        farm = Farm.objects.get(name="Wheat farm")
        get_default_farm_mock.return_value = farm

        self.workers_process.worker_runs.create(version=self.recognizer)
        self.assertFalse(self.workers_process.tasks.exists())

        self.client.force_login(self.user)

        with self.assertNumQueries(18):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.json()["id"], str(self.workers_process.id))
        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.state, State.Unscheduled)
        self.assertEqual(self.workers_process.farm, farm)

        self.assertEqual(activities_delay_mock.call_count, 1)
        self.assertEqual(activities_delay_mock.call_args, call(
            self.workers_process,
            user_id=self.user.id,
            description="Initializing worker activities for process Recognizer",
        ))

    @patch("arkindex.ponos.models.Farm.is_available", return_value=False)
    @patch("arkindex.process.serializers.imports.get_default_farm")
    def test_default_farm_guest(self, get_default_farm_mock, is_available_mock):
        get_default_farm_mock.return_value = Farm.objects.first()
        self.workers_process.worker_runs.create(version=self.recognizer)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "farm": ["You do not have access to this farm."],
        })
        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.state, State.Unscheduled)
        self.assertIsNone(self.workers_process.farm)
        self.assertEqual(get_default_farm_mock.call_count, 1)
        self.assertEqual(is_available_mock.call_count, 1)
        self.assertEqual(is_available_mock.call_args, call(self.user))

    @patch("arkindex.ponos.models.Farm.is_available", return_value=False)
    def test_farm_guest(self, is_available_mock):
        self.workers_process.worker_runs.create(version=self.recognizer)
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)}),
                {"farm": str(Farm.objects.first().id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "farm": ["You do not have access to this farm."],
        })
        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.state, State.Unscheduled)
        self.assertIsNone(self.workers_process.farm)
        self.assertEqual(is_available_mock.call_count, 1)
        self.assertEqual(is_available_mock.call_args, call(self.user))

    def test_wrong_farm_id(self):
        """
        Handle a payload with a wrong farm ID starting a process
        """
        self.client.force_login(self.user)
        wrong_farm_id = uuid4()

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)}),
                {"farm": str(wrong_farm_id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"farm": [f'Invalid pk "{wrong_farm_id}" - object does not exist.']})

    def test_params_validation(self):
        self.client.force_login(self.user)
        wrong_params_checks = [
            ({"chunks": 0}, {"chunks": ["Ensure this value is greater than or equal to 1."]}),
            ({"chunks": 20}, {"chunks": ["Ensure this value is less than or equal to 10."]}),
            ({"chunks": "max"}, {"chunks": ["A valid integer is required."]}),
        ]
        for (params, check) in wrong_params_checks:
            with self.subTest(**params), self.assertNumQueries(5):
                response = self.client.post(reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)}), params)
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), check)

    @override_settings(MAX_CHUNKS=42)
    def test_serializer_chunks_config(self):
        """
        StartProcess should restrict the chunks to `settings.MAX_CHUNKS`
        """
        self.workers_process.worker_runs.create(version=self.recognizer)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)}),
                {"chunks": 43},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "chunks": ["Ensure this value is less than or equal to 42."],
        })

    @patch("arkindex.project.triggers.process_tasks.initialize_activity.delay")
    def test_workers_parameters(self, activities_delay_mock):
        """
        It should be possible to pass chunks parameters when starting a workers process
        """
        run = self.workers_process.worker_runs.create(version=self.recognizer)

        self.client.force_login(self.user)
        with self.assertNumQueries(18):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)}),
                {"chunks": 3}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.workers_process.refresh_from_db()

        self.assertEqual(list(self.workers_process.tasks.order_by("slug").values_list("slug", flat=True)), [
            "initialisation",
            f"{run.task_slug}_1",
            f"{run.task_slug}_2",
            f"{run.task_slug}_3",
        ])

        self.assertEqual(activities_delay_mock.call_count, 1)
        self.assertEqual(activities_delay_mock.call_args, call(
            self.workers_process,
            user_id=self.user.id,
            description="Initializing worker activities for process Recognizer",
        ))

    def test_dataset_chunks(self):
        """
        It should be possible to pass chunks when starting a dataset process
        """
        test_set_1 = self.dataset1.sets.get(name="test")
        test_set_2 = self.dataset2.sets.get(name="test")
        ProcessDatasetSet.objects.create(process=self.dataset_process, set=test_set_1)
        ProcessDatasetSet.objects.create(process=self.dataset_process, set=test_set_2)
        run = self.dataset_process.worker_runs.create(version=self.recognizer)
        self.client.force_login(self.user)

        with self.assertNumQueries(12):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.dataset_process.id)}),
                {"chunks": 3}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(list(self.dataset_process.tasks.order_by("slug").values_list("slug", flat=True)), [
            f"{run.task_slug}_1",
            f"{run.task_slug}_2",
            f"{run.task_slug}_3",
        ])

    @patch("arkindex.process.models.Process.worker_runs")
    @patch("arkindex.project.triggers.process_tasks.initialize_activity.delay")
    def test_options_requires_workers(self, activities_delay_mock, worker_runs_mock):
        """
        The cache option requires the process to have workers attached
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)}),
                {"use_cache": "true"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "__all__": ["The process must have worker runs to be started."]
        })
        self.workers_process.refresh_from_db()
        self.assertFalse(self.workers_process.use_cache)
        self.assertEqual(self.workers_process.activity_state, ActivityState.Disabled)
        self.assertFalse(activities_delay_mock.called)

    @patch("arkindex.project.triggers.process_tasks.initialize_activity.delay")
    def test_options(self, activities_delay_mock):
        """
        The cache option can be triggered when starting a process, and worker activities cannot be disabled
        """
        self.assertFalse(self.workers_process.use_cache)
        self.assertEqual(self.workers_process.activity_state, ActivityState.Disabled)
        self.workers_process.worker_runs.create(version=self.version_gpu)

        self.client.force_login(self.user)

        with self.assertNumQueries(18):
            response = self.client.post(
                reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)}),
                {"use_cache": "true", "worker_activity": "false"},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())

        self.workers_process.refresh_from_db()
        self.assertTrue(self.workers_process.use_cache)
        # The process activity state is set to Pending while activities are created in an async job
        self.assertEqual(self.workers_process.activity_state, ActivityState.Pending)
        self.assertEqual(activities_delay_mock.call_count, 1)
        self.assertEqual(activities_delay_mock.call_args, call(
            self.workers_process,
            user_id=self.user.id,
            description="Initializing worker activities for process Worker requiring a GPU",
        ))

    @patch("arkindex.project.triggers.process_tasks.initialize_activity.delay")
    def test_duplicated_worker_version(self, activities_delay_mock):
        """
        A WorkerRun depending on another WorkerRun of the same WorkerVersion should be supported
        """
        run_1 = self.workers_process.worker_runs.create(
            version=self.recognizer,
            configuration=self.recognizer.worker.configurations.create(
                name="some_config",
                configuration={"a": "b"},
            ),
        )
        run_2 = self.workers_process.worker_runs.create(
            version=self.recognizer,
            parents=[run_1.id],
        )
        self.assertNotEqual(run_1.task_slug, run_2.task_slug)

        self.client.force_login(self.user)
        with self.assertNumQueries(18):
            response = self.client.post(reverse("api:process-start", kwargs={"pk": str(self.workers_process.id)}))
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(self.workers_process.tasks.count(), 3)
        init_task = self.workers_process.tasks.get(slug="initialisation")
        task_1 = self.workers_process.tasks.get(slug=run_1.task_slug)
        task_2 = self.workers_process.tasks.get(slug=run_2.task_slug)

        self.assertQuerySetEqual(task_1.parents.all(), [init_task])
        self.assertQuerySetEqual(task_2.parents.all(), [task_1])

        self.assertEqual(activities_delay_mock.call_count, 1)
        self.assertEqual(activities_delay_mock.call_args, call(
            self.workers_process,
            user_id=self.user.id,
            description="Initializing worker activities for process Recognizer, Recognizer",
        ))
