from datetime import datetime, timezone
from unittest.mock import patch

from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Agent, Farm, State
from arkindex.process.models import ProcessMode, WorkerRun, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestWorkerRunsDelete(FixtureAPITestCase):
    """
    Test worker runs delete endpoint
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.local_process = cls.user.processes.get(mode=ProcessMode.Local)
        cls.farm = Farm.objects.first()
        cls.process_1 = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            farm=cls.farm,
        )
        cls.version_1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_1 = cls.version_1.worker
        cls.version_2 = WorkerVersion.objects.get(worker__slug="dla")
        cls.run_1 = cls.process_1.worker_runs.create(version=cls.version_1, parents=[])

        cls.agent = Agent.objects.create(
            farm=cls.farm,
            hostname="claude",
            cpu_cores=42,
            cpu_frequency=1e15,
            ram_total=99e9,
            last_ping=datetime.now(timezone.utc),
        )
        # Add custom attributes to make the agent usable as an authenticated user
        cls.agent.is_agent = True
        cls.agent.is_anonymous = False

    def test_delete_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_invalid_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": "12341234-1234-1234-1234-123412341234"})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_delete_no_worker_execution_right(self):
        """
        A user can delete a worker run with no right on its worker but an admin access to the process project
        """
        self.worker_1.memberships.update(level=Role.Guest.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(WorkerRun.DoesNotExist):
            self.run_1.refresh_from_db()

    def test_delete_local(self):
        """
        A user cannot delete a worker run on a local process
        """
        run = self.local_process.worker_runs.create(version=self.version_1, parents=[])
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), ["WorkerRuns can only be deleted from Workers or Dataset processes."])

    def test_delete(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(WorkerRun.DoesNotExist):
            self.run_1.refresh_from_db()

    def test_delete_with_children(self):
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=2,
            configuration={"test": "test2"}
        )
        version_3 = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=3,
            configuration={"test": "test3"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[self.run_1.id],
        )
        run_3 = self.process_1.worker_runs.create(
            version=version_3,
            parents=[self.run_1.id, run_2.id],
        )

        self.assertTrue(self.run_1.id in run_2.parents)
        self.assertTrue(self.run_1.id in run_3.parents)
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(WorkerRun.DoesNotExist):
            self.run_1.refresh_from_db()
        run_2.refresh_from_db()
        run_3.refresh_from_db()
        self.assertEqual(run_2.parents, [])
        self.assertEqual(run_3.parents, [run_2.id])

    def test_delete_started_process(self):
        """
        A user shouldn't be able to delete the worker run of a started process
        """
        self.client.force_login(self.user)
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process_1.run()
        self.process_1.tasks.update(state=State.Running)

        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), ["WorkerRuns cannot be deleted from a process that has already started."])

    def test_delete_agent(self):
        """
        Ponos agents cannot delete WorkerRuns, even when they can access them
        """
        self.process_1.tasks.create(run=0, depth=0, slug="something", agent=self.agent)

        # Agent auth is not implemented in CE
        self.client.force_authenticate(user=self.agent)
        with self.assertNumQueries(1):
            response = self.client.delete(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {
            "detail": "You do not have an admin access to this process.",
        })

    def test_wrong_process_mode(self):
        self.client.force_login(self.user)

        for mode in set(ProcessMode) - {ProcessMode.Workers, ProcessMode.Dataset, ProcessMode.Local}:
            with self.subTest(mode=mode):
                self.process_1.mode = mode
                self.process_1.save()

                with self.assertNumQueries(3):
                    response = self.client.delete(
                        reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}))
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), ["WorkerRuns can only be deleted from Workers or Dataset processes."])
