import uuid
from datetime import datetime, timezone
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Agent, Farm
from arkindex.process.models import FeatureUsage, ProcessMode, WorkerRun, WorkerVersion, WorkerVersionState
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersion, ModelVersionState
from arkindex.users.models import Role


class TestWorkerRunsPartialUpdate(FixtureAPITestCase):
    """
    Test worker runs partial update endpoint
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.local_process = cls.user.processes.get(mode=ProcessMode.Local)
        cls.farm = Farm.objects.first()
        cls.process_1 = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            farm=cls.farm,
        )
        cls.version_1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_1 = cls.version_1.worker
        cls.version_2 = WorkerVersion.objects.get(worker__slug="dla")
        cls.run_1 = cls.process_1.worker_runs.create(version=cls.version_1, parents=[])
        cls.configuration_1 = cls.worker_1.configurations.create(name="My config", configuration={"key": "value"})
        worker_version = WorkerVersion.objects.exclude(worker=cls.version_1.worker).first()
        cls.configuration_2 = worker_version.worker.configurations.create(name="Config")
        cls.process_2 = cls.corpus.processes.create(creator=cls.user, mode=ProcessMode.Workers)
        cls.version_gpu_required = WorkerVersion.objects.create(
            worker=cls.worker_1,
            configuration={"pilot": "ikari shinji"},
            state=WorkerVersionState.Available,
            model_usage=FeatureUsage.Disabled,
            docker_image_iid="registry.nerv.co.jp/neon-genesis:evangelion",
            gpu_usage=FeatureUsage.Required,
            version=3
        )
        cls.version_gpu_supported = WorkerVersion.objects.create(
            worker=cls.worker_1,
            configuration={"pilot": "ayanami rei"},
            state=WorkerVersionState.Available,
            model_usage=FeatureUsage.Disabled,
            docker_image_iid="registry.nerv.co.jp/neon-genesis:evangelion",
            gpu_usage=FeatureUsage.Supported,
            version=4
        )
        cls.version_gpu_disabled = WorkerVersion.objects.create(
            worker=cls.worker_1,
            configuration={"pilot": "soryu asuka langley"},
            state=WorkerVersionState.Available,
            model_usage=FeatureUsage.Disabled,
            docker_image_iid="registry.nerv.co.jp/neon-genesis:evangelion",
            gpu_usage=FeatureUsage.Disabled,
            version=5
        )

        # Model and Model version setup
        cls.model_1 = Model.objects.create(name="My model")
        cls.model_1.memberships.create(user=cls.user, level=Role.Contributor.value)
        cls.model_version_1 = ModelVersion.objects.create(
            model=cls.model_1,
            state=ModelVersionState.Available,
            size=8,
            hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            archive_hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
        )

        cls.model_2 = Model.objects.create(name="Their model")
        cls.model_2.memberships.create(user=cls.user, level=Role.Guest.value)
        cls.model_version_2 = cls.model_2.versions.create(
            state=ModelVersionState.Available,
            tag="blah",
            size=8,
            hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            archive_hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
        )

        cls.model_3 = Model.objects.create(name="Our model", public=True)
        cls.model_version_3 = cls.model_3.versions.create(
            state=ModelVersionState.Available,
            tag="blah",
            size=8,
            hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            archive_hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
        )

        cls.agent = Agent.objects.create(
            farm=cls.farm,
            hostname="claude",
            cpu_cores=42,
            cpu_frequency=1e15,
            ram_total=99e9,
            last_ping=datetime.now(timezone.utc),
        )
        # Add custom attributes to make the agent usable as an authenticated user
        cls.agent.is_agent = True
        cls.agent.is_anonymous = False

    def test_partial_update_requires_login(self):
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )

        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": [str(run_2.id)],
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.get_max_level", return_value=Role.Contributor.value)
    def test_partial_update_no_project_admin_right(self, get_max_level_mock):
        """
        A user cannot update a worker run if they have no admin access on its process project
        """
        self.corpus.memberships.filter(user=self.user).update(level=Role.Contributor.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have an admin access to this process."})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    def test_partial_update_invalid_id(self):
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )

        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": "12341234-1234-1234-1234-123412341234"}),
                data={
                    "parents": [str(run_2.id)],
                },
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_partial_update_local(self):
        """
        A user cannot update a worker run on a local process
        """
        run = self.local_process.worker_runs.create(version=self.version_1, parents=[])
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                data={
                    "parents": []
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "process_id": ["WorkerRuns can only be created or updated on Workers or Dataset processes."],
        })

    def test_partial_update_nonexistent_parent(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": ["12341234-1234-1234-1234-123412341234"],
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), [
            f"Can't add or update WorkerRun {self.run_1.id} because parents field isn't properly defined. It can be either because"
            " one or several UUIDs don't refer to existing WorkerRuns or either because listed WorkerRuns doesn't belong to the"
            " same Process than this WorkerRun."
        ])

    def test_partial_update_process_id(self):
        """
        Process field cannot be updated
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "process_id": str(self.process_2.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image_iid": self.version_1.docker_image_iid,
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision_url": None,
                "version": 1,
                "tag": None,
                "branch": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer",
                    "description": "",
                    "repository_url": None,
                    "archived": False,
                }
            },
            "use_gpu": False,
            "summary": "Worker Recognizer @ version 1",
        })
        self.run_1.refresh_from_db()
        self.assertEqual(self.run_1.process.id, self.process_1.id)

    def test_partial_update_worker_version_id(self):
        """
        Version field cannot be updated
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "worker_version_id": str(self.version_2.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image_iid": self.version_1.docker_image_iid,
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision_url": None,
                "version": 1,
                "tag": None,
                "branch": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer",
                    "description": "",
                    "repository_url": None,
                    "archived": False,
                }
            },
            "use_gpu": False,
            "summary": "Worker Recognizer @ version 1",
        })
        self.run_1.refresh_from_db()
        self.assertNotEqual(self.run_1.version_id, self.version_2.id)

    def test_partial_update_configuration(self):
        self.client.force_login(self.user)
        self.assertEqual(self.run_1.configuration, None)
        self.assertEqual(self.run_1.summary, "Worker Recognizer @ version 1")

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": self.run_1.id}),
                data={
                    "configuration_id": str(self.configuration_1.id)
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.run_1.refresh_from_db()
        self.assertEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": {
                "id": str(self.configuration_1.id),
                "archived": False,
                "configuration": {"key": "value"},
                "name": "My config"
            },
            "model_version": None,
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image_iid": self.version_1.docker_image_iid,
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision_url": None,
                "version": 1,
                "tag": None,
                "branch": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer",
                    "description": "",
                    "repository_url": None,
                    "archived": False,
                }
            },
            "use_gpu": False,
            "summary": "Worker Recognizer @ version 1 using configuration 'My config'",
        })
        self.assertEqual(self.run_1.configuration.id, self.configuration_1.id)
        self.assertEqual(self.run_1.summary, "Worker Recognizer @ version 1 using configuration 'My config'")

    def test_partial_update_invalid_configuration(self):
        self.client.force_login(self.user)
        self.assertEqual(self.run_1.configuration, None)

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": self.run_1.id}),
                data={"configuration_id": str(self.configuration_2.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"configuration_id": ["The configuration must be part of the same worker."]})

    def test_partial_update_process_already_started(self):
        """
        Update dependencies of a worker run is not possible once the process is started
        """
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.process_1.run()
        self.assertTrue(self.process_1.tasks.exists())
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": [],
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "process_id": ["WorkerRuns cannot be added or updated on processes that have already started."],
        })

    def test_partial_update_model_version_not_allowed(self):
        """
        The model_version UUID is not allowed when the related version doesn't allow model_usage
        """
        self.client.force_login(self.user)
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Disabled
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                data={
                    "model_version_id": str(self.model_version_1.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": ["This worker version does not support models."]
        })

    def test_partial_update_unknown_model_version(self):
        """
        Cannot use a model version id that doesn't exist
        """
        self.client.force_login(self.user)
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )
        random_model_version_uuid = str(uuid.uuid4())

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                data={
                    "model_version_id": random_model_version_uuid,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": [f'Invalid pk "{random_model_version_uuid}" - object does not exist.']
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=ModelVersion.objects.none())
    def test_partial_update_model_version_no_access(self, filter_rights_mock):
        """
        Cannot update a worker run with a model_version UUID, when you don't have access to the model version
        """
        self.client.force_login(self.user)
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )

        # Create a model version, the user has no access to
        model_no_access = Model.objects.create(name="Secret model")
        model_version_no_access = ModelVersion.objects.create(model=model_no_access, state=ModelVersionState.Available, size=8, hash="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", archive_hash="bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb")

        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                data={
                    "model_version_id": str(model_version_no_access.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": [f'Invalid pk "{model_version_no_access.id}" - object does not exist.'],
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Model, Role.Guest.value),
            call(self.user, Model, Role.Contributor.value),
        ])

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_partial_update_model_version_guest(self, filter_rights_mock):
        """
        Cannot update a worker run with a model_version when you only have guest access to the model,
        and the model version has no tag or is not available
        """
        self.client.force_login(self.user)
        version_no_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_no_model,
            parents=[],
        )

        def filter_rights(user, model, level):
            """
            The filter_rights mock needs to return nothing when called for contributor access,
            and the models we will test on when called for guest access
            """
            if level == Role.Guest.value:
                return Model.objects.filter(id__in=(self.model_2.id, self.model_3.id))
            return Model.objects.none()

        filter_rights_mock.side_effect = filter_rights

        cases = [
            # On a model with a membership giving guest access
            (self.model_version_2, None, ModelVersionState.Created),
            (self.model_version_2, None, ModelVersionState.Error),
            (self.model_version_2, None, ModelVersionState.Available),
            (self.model_version_2, "blah", ModelVersionState.Created),
            (self.model_version_2, "blah", ModelVersionState.Error),
            # On a public model with no membership
            (self.model_version_3, None, ModelVersionState.Created),
            (self.model_version_3, None, ModelVersionState.Error),
            (self.model_version_3, None, ModelVersionState.Available),
            (self.model_version_3, "blah", ModelVersionState.Created),
            (self.model_version_3, "blah", ModelVersionState.Error),
        ]

        for model_version, tag, state in cases:
            filter_rights_mock.reset_mock()
            with self.subTest(model_version=model_version, tag=tag, state=state):
                model_version.tag = tag
                model_version.state = state
                model_version.save()

                with self.assertNumQueries(4):
                    response = self.client.patch(
                        reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                        data={
                            "model_version_id": str(model_version.id),
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "model_version_id": [f'Invalid pk "{model_version.id}" - object does not exist.'],
                })

                self.assertListEqual(filter_rights_mock.call_args_list, [
                    call(self.user, Model, Role.Guest.value),
                    call(self.user, Model, Role.Contributor.value),
                ])

    def test_partial_update_model_version_unavailable(self):
        self.client.force_login(self.user)
        version = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run = self.process_1.worker_runs.create(
            version=version,
            parents=[],
        )
        self.model_version_1.state = ModelVersionState.Error
        self.model_version_1.save()

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                data={
                    "model_version_id": str(self.model_version_1.id),
                    "parents": []
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": ["This ModelVersion is not in an Available state."],
        })

    def test_partial_update_model_archived(self):
        self.client.force_login(self.user)
        version = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run = self.process_1.worker_runs.create(version=version)
        self.model_1.archived = datetime.now(timezone.utc)
        self.model_1.save()

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                data={
                    "model_version_id": str(self.model_version_1.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model_version_id": ["This ModelVersion is part of an archived model."],
        })

    def test_partial_update_model_version(self):
        """
        Update the worker run by adding a model_version with a worker version that supports it
        """
        self.client.force_login(self.user)
        version_with_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run = self.process_1.worker_runs.create(
            version=version_with_model,
            parents=[],
        )
        self.assertIsNone(run.model_version_id)
        self.assertEqual(run.summary, "Worker Recognizer @ version 2")

        model_versions = [
            # Version on a model with contributor access
            self.model_version_1,
            # Available version with tag on a model with guest access
            self.model_version_2,
            # Available version with tag on a public model
            self.model_version_3,
        ]
        for model_version in model_versions:
            with self.subTest(model_version=model_version):
                with self.assertNumQueries(6):
                    response = self.client.patch(
                        reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                        data={
                            "model_version_id": str(model_version.id),
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                run.refresh_from_db()
                self.assertEqual(response.json(), {
                    "id": str(run.id),
                    "configuration": None,
                    "model_version": {
                        "id": str(model_version.id),
                        "configuration": {},
                        "model": {
                            "id": str(model_version.model.id),
                            "name": model_version.model.name
                        },
                        "size": 8,
                        "state": "available",
                        "tag": model_version.tag,
                    },
                    "parents": [],
                    "process": {
                        "id": str(self.process_1.id),
                        "activity_state": "disabled",
                        "corpus": str(self.corpus.id),
                        "chunks": 1,
                        "mode": "workers",
                        "name": None,
                        "state": "unscheduled",
                        "use_cache": False,
                        "element": None,
                        "element_type": None,
                        "folder_type": None,
                        "prefix": None,
                    },
                    "worker_version": {
                        "id": str(version_with_model.id),
                        "configuration": {"test": "test2"},
                        "docker_image_iid": None,
                        "gpu_usage": "disabled",
                        "model_usage": FeatureUsage.Required.value,
                        "revision_url": None,
                        "version": version_with_model.version,
                        "tag": None,
                        "branch": None,
                        "created": version_with_model.created.isoformat().replace("+00:00", "Z"),
                        "state": "created",
                        "worker": {
                            "id": str(self.worker_1.id),
                            "name": "Recognizer",
                            "slug": "reco",
                            "type": "recognizer",
                            "description": "",
                            "repository_url": None,
                            "archived": False,
                        }
                    },
                    "use_gpu": False,
                    "summary": f"Worker Recognizer @ version 2 with model {model_version.model.name} @ {str(model_version.id)[:6]}",
                })
                self.assertEqual(run.model_version_id, model_version.id)
                self.assertEqual(run.summary, f"Worker Recognizer @ version 2 with model {model_version.model.name} @ {str(model_version.id)[:6]}")

    def test_partial_update_model_version_with_configuration(self):
        """
        Updating the worker run by adding a model_version with a worker version
        doesn't erase previously loaded worker configuration
        """
        self.client.force_login(self.user)
        version_with_model = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=2,
            configuration={"test": "test2"},
            model_usage=FeatureUsage.Required
        )
        run = self.process_1.worker_runs.create(
            version=version_with_model,
            parents=[],
            configuration=self.configuration_1
        )
        self.assertEqual(run.model_version_id, None)

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                data={
                    "model_version_id": str(self.model_version_1.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        run.refresh_from_db()
        self.assertEqual(response.json(), {
            "id": str(run.id),
            "configuration": {
                "archived": False,
                "configuration": {"key": "value"},
                "id": str(self.configuration_1.id),
                "name": "My config"
            },
            "model_version": {
                "id": str(self.model_version_1.id),
                "configuration": {},
                "model": {
                    "id": str(self.model_1.id),
                    "name": "My model"
                },
                "size": 8,
                "state": "available",
                "tag": None
            },
            "parents": [],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(version_with_model.id),
                "configuration": {"test": "test2"},
                "docker_image_iid": None,
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Required.value,
                "revision_url": None,
                "version": version_with_model.version,
                "tag": None,
                "branch": None,
                "created": version_with_model.created.isoformat().replace("+00:00", "Z"),
                "state": "created",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer",
                    "description": "",
                    "repository_url": None,
                    "archived": False,
                }
            },
            "use_gpu": False,
            "summary": f"Worker Recognizer @ version 2 with model My model @ {str(self.model_version_1.id)[:6]} using configuration 'My config'",
        })
        self.assertEqual(run.model_version_id, self.model_version_1.id)
        self.assertEqual(run.configuration_id, self.configuration_1.id)

    def test_partial_update(self):
        version_2 = WorkerVersion.objects.create(
            worker=self.worker_1,
            version=2,
            configuration={"test": "test2"}
        )
        run_2 = self.process_1.worker_runs.create(
            version=version_2,
            parents=[],
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                data={
                    "parents": [str(run_2.id)],
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.run_1.refresh_from_db()
        self.assertDictEqual(response.json(), {
            "id": str(self.run_1.id),
            "configuration": None,
            "model_version": None,
            "parents": [str(run_2.id)],
            "process": {
                "id": str(self.process_1.id),
                "activity_state": "disabled",
                "corpus": str(self.corpus.id),
                "chunks": 1,
                "mode": "workers",
                "name": None,
                "state": "unscheduled",
                "use_cache": False,
                "element": None,
                "element_type": None,
                "folder_type": None,
                "prefix": None,
            },
            "worker_version": {
                "id": str(self.version_1.id),
                "configuration": {"test": 42},
                "docker_image_iid": self.version_1.docker_image_iid,
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
                "revision_url": None,
                "version": 1,
                "tag": None,
                "branch": None,
                "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                "state": "available",
                "worker": {
                    "id": str(self.worker_1.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "type": "recognizer",
                    "description": "",
                    "repository_url": None,
                    "archived": False,
                }
            },
            "use_gpu": False,
            "summary": "Worker Recognizer @ version 1",
        })

    def test_partial_update_unique(self):
        self.client.force_login(self.user)
        self.version_1.model_usage = FeatureUsage.Required
        self.version_1.save()
        cases = [
            (None, None),
            (None, self.configuration_1),
            (self.model_version_1, None),
            (self.model_version_1, self.configuration_1),
        ]
        for model_version, configuration in cases:
            with self.subTest(model_version=model_version, configuration=configuration):
                # Erase any previous failures
                self.process_1.worker_runs.exclude(id=self.run_1.id).delete()

                self.run_1.model_version = model_version
                self.run_1.configuration = configuration
                self.run_1.save()

                # Ensure the other run has different values before updating to avoid conflicts
                run_2 = self.process_1.worker_runs.create(
                    version=self.version_1,
                    model_version=None if model_version else self.model_version_1,
                    configuration=None if configuration else self.configuration_1,
                )

                # Having a model version or a configuration adds one query for each
                query_count = 4 + bool(model_version) + bool(configuration)

                with self.assertNumQueries(query_count):
                    response = self.client.patch(
                        reverse("api:worker-run-details", kwargs={"pk": str(run_2.id)}),
                        data={
                            # Update the second worker run to the first worker run's values to cause a conflict
                            "model_version_id": str(model_version.id) if model_version else None,
                            "configuration_id": str(configuration.id) if configuration else None,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "__all__": ["A WorkerRun already exists on this process with the selected worker version, model version and configuration."],
                })

    def test_partial_update_agent(self):
        """
        Ponos agents cannot update WorkerRuns, even when they can access them
        """
        self.process_1.tasks.create(run=0, depth=0, slug="something", agent=self.agent)

        # Agent auth is not implemented in CE
        self.client.force_authenticate(user=self.agent)
        with self.assertNumQueries(1):
            response = self.client.patch(
                reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {
            "detail": "You do not have an admin access to this process.",
        })

    def test_update_use_gpu_errors(self):
        self.client.force_login(self.user)

        cases = [
            (self.version_gpu_disabled, True, "This worker version does not support GPU usage."),
            (self.version_gpu_required, False, "This worker version requires GPU usage.")
        ]

        for worker_version, use_gpu, error_message in cases:
            with self.subTest(worker_version=worker_version, use_gpu=use_gpu, error_message=error_message):
                run = WorkerRun.objects.create(
                    process=self.process_1,
                    version=worker_version,
                    parents=[]
                )
                self.assertEqual(run.use_gpu, True if worker_version.gpu_usage == FeatureUsage.Required else False)
                with self.assertNumQueries(3):
                    response = self.client.patch(
                        reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                        data={"use_gpu": use_gpu}
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                    self.assertDictEqual(response.json(), {"use_gpu": [error_message]})

                run.refresh_from_db()
                self.assertEqual(run.use_gpu, True if worker_version.gpu_usage == FeatureUsage.Required else False)

    def test_update_use_gpu(self):
        self.client.force_login(self.user)

        cases = [
            (self.version_gpu_disabled, False),
            (self.version_gpu_required, True),
            (self.version_gpu_supported, False),
            (self.version_gpu_supported, True)
        ]

        for worker_version, use_gpu in cases:
            with self.subTest(worker_version=worker_version, use_gpu=use_gpu):
                self.process_1.worker_runs.all().delete()
                run = WorkerRun.objects.create(
                    process=self.process_1,
                    version=worker_version,
                    parents=[]
                )
                self.assertEqual(run.use_gpu, True if worker_version.gpu_usage == FeatureUsage.Required else False)

                with self.assertNumQueries(5):
                    response = self.client.patch(
                        reverse("api:worker-run-details", kwargs={"pk": str(run.id)}),
                        data={"use_gpu": use_gpu}
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                data = response.json()
                self.assertDictEqual(data, {
                    "id": str(run.id),
                    "worker_version": {
                        "id": str(worker_version.id),
                        "configuration": worker_version.configuration,
                        "docker_image_iid": "registry.nerv.co.jp/neon-genesis:evangelion",
                        "gpu_usage": worker_version.gpu_usage.value,
                        "model_usage": FeatureUsage.Disabled.value,
                        "revision_url": None,
                        "version": worker_version.version,
                        "tag": None,
                        "branch": None,
                        "created": worker_version.created.isoformat().replace("+00:00", "Z"),
                        "state": "available",
                        "worker": {
                            "id": str(self.worker_1.id),
                            "name": "Recognizer",
                            "slug": "reco",
                            "type": "recognizer",
                            "description": "",
                            "repository_url": None,
                            "archived": False,
                        }
                    },
                    "parents": [],
                    "model_version": None,
                    "configuration": None,
                    "process": {
                        "id": str(self.process_1.id),
                        "activity_state": "disabled",
                        "corpus": str(self.corpus.id),
                        "element": None,
                        "element_type": None,
                        "folder_type": None,
                        "prefix": None,
                        "chunks": 1,
                        "mode": "workers",
                        "name": None,
                        "state": "unscheduled",
                        "use_cache": False,
                    },
                    "summary": f"Worker Recognizer @ version {worker_version.version}",
                    "use_gpu": use_gpu
                })
                run.refresh_from_db()
                self.assertEqual(run.use_gpu, use_gpu)

    def test_wrong_process_mode(self):
        self.client.force_login(self.user)

        for mode in set(ProcessMode) - {ProcessMode.Workers, ProcessMode.Dataset, ProcessMode.Local}:
            with self.subTest(mode=mode):
                self.process_1.mode = mode
                self.process_1.save()

                with self.assertNumQueries(4):
                    response = self.client.patch(
                        reverse("api:worker-run-details", kwargs={"pk": str(self.run_1.id)}),
                        data={
                            "use_gpu": False,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), {"process_id": ["WorkerRuns can only be created or updated on Workers or Dataset processes."]})
