from datetime import datetime, timezone
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Farm
from arkindex.process.models import FeatureUsage, ProcessMode, Worker, WorkerVersion, WorkerVersionState
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model
from arkindex.users.models import Role


class TestWorkerVersionCreate(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_reco = Worker.objects.get(slug="reco")
        cls.worker_dla = Worker.objects.get(slug="dla")

        cls.version_1 = cls.worker_reco.versions.get()
        cls.version_2 = cls.worker_dla.versions.get()

        farm = Farm.objects.first()
        process = cls.corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=cls.user,
            farm=farm,
            name="test_process"
        )
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            process.run()
        cls.task = process.tasks.get()

        cls.model = Model.objects.create(
            name="Generic model",
            public=False,
        )

    def test_create_non_existing_worker(self):
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": "12341234-1234-1234-1234-123412341234"}),
                data={"configuration": {"test": "test2"}},
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertDictEqual(response.json(), {"detail": "No Worker matches the given query."})

    def test_create_available_requires_docker_image(self):
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "configuration": {"test": "test2"},
                    "state": "available",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "state": [
                "A worker version must have a `docker_image_iid` set when updated to available."
            ]
        })

    def test_create_invalid_docker_image(self):
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "docker_image_iid": "http://seele.co.jp/",
                    "configuration": {"test": "test2"},
                    "state": "available",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "docker_image_iid": [
                "Repository name cannot contain a scheme (http://seele.co.jp/)"
            ]
        })

    def test_create_empty(self):
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"configuration": ["This field is required."]})

    def test_create_archived(self):
        self.worker_reco.archived = datetime.now(timezone.utc)
        self.worker_reco.save()

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "configuration": {},
                    "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01"
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "worker": ["This worker is archived."],
        })

    def test_create_null_revision_url_requires_null_worker_repository_url(self):
        self.worker_reco.repository_url = "https://gitlab.com/NERV/eva"
        self.worker_reco.save()
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"configuration": {"test": "test2"}, "model_usage": FeatureUsage.Required.value},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "revision_url": ["A revision url is required when creating a version for a worker linked to a repository."],
        })

    def test_create_null_revision_url_forbidden_task_auth(self):
        """
        Ponos Task auth cannot create a version on a worker that is not linked to a repository.
        """
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"configuration": {"test": "test2"}, "model_usage": FeatureUsage.Required.value},
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "revision_url": ["This field is required when creating a version through Task authentication."]
        })

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Contributor.value)
    def test_create_user_auth_requires_admin_access_to_worker(self, max_level_mock):
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"configuration": {"test": "val"}, "model_usage": FeatureUsage.Required.value},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have an admin access to this worker."})

        self.assertEqual(max_level_mock.call_count, 1)
        self.assertEqual(max_level_mock.call_args, call(self.user, self.worker_reco))

    def test_create_user_auth_with_worker_repository_url_ok(self):
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)
        self.worker_dla.repository_url = "https://gitlab.com/NERV/eva"
        self.worker_dla.save()

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
                data={"revision_url": "https://gitlab.com/NERV/eva/commit/eva-01", "configuration": {"test": "test2"}, "model_usage": FeatureUsage.Required.value},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        new_version = self.worker_dla.versions.get(id=data["id"])
        self.assertDictEqual(response.json(), {
            "id": str(new_version.id),
            "configuration": {"test": "test2"},
            "created": new_version.created.isoformat().replace("+00:00", "Z"),
            "docker_image_iid": None,
            "gpu_usage": FeatureUsage.Disabled.value,
            "model_usage": FeatureUsage.Required.value,
            "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
            "state": WorkerVersionState.Created.value,
            "version": None,
            "tag": None,
            "branch": None,
            "worker": {
                "id": str(self.worker_dla.id),
                "name": "Document layout analyser",
                "slug": "dla",
                "type": "dla",
                "description": "",
                "archived": False,
                "repository_url": "https://gitlab.com/NERV/eva",
            },
        })

    def test_create_null_revision_url(self):
        """
        A worker version can be created with no revision_url through user authentication
        """
        self.worker_reco.versions.create(version=41, configuration={})
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"configuration": {"test": "val"}, "model_usage": FeatureUsage.Required.value},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        new_version = self.worker_reco.versions.get(id=data["id"])
        self.assertDictEqual(response.json(), {
            "id": str(new_version.id),
            "configuration": {"test": "val"},
            "created": new_version.created.isoformat().replace("+00:00", "Z"),
            "docker_image_iid": None,
            "gpu_usage": FeatureUsage.Disabled.value,
            "model_usage": FeatureUsage.Required.value,
            "revision_url": None,
            "state": WorkerVersionState.Created.value,
            "version": 42,
            "tag": None,
            "branch": None,
            "worker": {
                "id": str(self.worker_reco.id),
                "name": "Recognizer",
                "slug": "reco",
                "type": "recognizer",
                "description": "",
                "archived": False,
                "repository_url": None,
            },
        })

    def test_create_configuration_wrong_type(self):
        """
        Configuration body must be an object
        """
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"configuration": "test"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": ['Expected a dictionary of items but got type "str".']})

    def test_create_requires_permission(self):
        self.assertFalse(self.user.can_create_worker_version)
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
            data={
                "configuration": {"test": "test2"},
                "model_usage": FeatureUsage.Required.value,
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "You do not have permission to create workers or worker versions."})
        self.user.can_create_worker_version = True
        self.user.save()
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
            data={
                "configuration": {"test": "test2"},
                "model_usage": FeatureUsage.Required.value,
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_task_auth(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
            data={
                "configuration": {"test": "test2"},
                "model_usage": FeatureUsage.Required.value,
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        self.assertNotEqual(data["id"], str(self.version_1.id))
        self.assertEqual(data["configuration"], {"test": "test2"})
        self.assertEqual(data["version"], None)
        self.assertEqual(data["revision_url"], "https://gitlab.com/NERV/eva/commit/eva-01")
        self.assertEqual(data["state"], "created")
        self.assertEqual(data["gpu_usage"], "disabled")
        self.assertEqual(data["model_usage"], FeatureUsage.Required.value)

    def test_create_duplicate_revision_url(self):
        self.version_1.revision_url = "https://gitlab.com/NERV/eva/commit/eva-01"
        self.version_1.version = None
        self.version_1.save()
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
            data={
                "configuration": {"test": "test2"},
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"revision_url": ["A version already exists for this worker with this revision_url."]})

    def test_create_with_tag(self):
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "configuration": {"test": "test2"},
                    "model_usage": FeatureUsage.Required.value,
                    "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                    "tag": "eva-01",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        self.assertNotEqual(data["id"], str(self.version_1.id))
        self.assertEqual(data["configuration"], {"test": "test2"})
        self.assertEqual(data["version"], None)
        self.assertEqual(data["revision_url"], "https://gitlab.com/NERV/eva/commit/eva-01")
        self.assertEqual(data["state"], "created")
        self.assertEqual(data["gpu_usage"], "disabled")
        self.assertEqual(data["model_usage"], FeatureUsage.Required.value)
        self.assertEqual(data["tag"], "eva-01")

    def test_create_unique_tag(self):
        self.version_1.tag = "eva-01"
        self.version_1.save()

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"configuration": {"test": "test2"}, "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01", "tag": "eva-01"},
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"tag": ["A version already exists for this worker with this tag."]})

    def test_create_unassign_branch(self):
        self.version_1.branch = "operation-yashima"
        self.version_1.save()
        self.version_2.branch = "operation-yashima"
        self.version_2.save()

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "configuration": {"test": "test2"},
                    "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                    "branch": "operation-yashima",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        new_version = WorkerVersion.objects.get(id=response.json()["id"])
        self.assertEqual(new_version.branch, "operation-yashima")

        self.version_1.refresh_from_db()
        self.assertIsNone(self.version_1.branch)

        self.version_2.refresh_from_db()
        self.assertEqual(self.version_2.branch, "operation-yashima")

    def test_create_empty_tag_branch(self):
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "configuration": {"test": "test2"},
                    "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                    "tag": "",
                    "branch": "",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "tag": ["This field may not be blank."],
            "branch": ["This field may not be blank."],
        })

    def test_create_wrong_gpu_usage(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
            data={"configuration": {"test": "test2"}, "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01", "gpu_usage": "not_supported"},
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_no_user_configuration_ok(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {"beep": "boop"},
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_valid_user_configuration(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {"title": "Demo Integer", "type": "int", "required": True, "default": 1},
                        "demo_boolean": {"title": "Demo Boolean", "type": "bool", "required": False, "default": True},
                    }
                },
                "gpu_usage": "disabled",
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_integer": {
                    "title": "Demo Integer",
                    "type": "int",
                    "required": True,
                    "default": 1
                },
                "demo_boolean": {
                    "title": "Demo Boolean",
                    "type": "bool",
                    "required": False,
                    "default": True
                }
            }
        })

    def test_create_valid_user_configuration_dict(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_dict": {"title": "Demo Dict", "type": "dict", "required": True, "default": {"a": "b", "c": "d"}},
                    }
                },
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_dict": {
                    "title": "Demo Dict",
                    "type": "dict",
                    "required": True,
                    "default": {"a": "b", "c": "d"}
                }
            }
        })

    def test_create_user_configuration_dict_strings_only(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_dict": {"title": "Demo Dict", "type": "dict", "required": True, "default": {"a": ["12", "13"], "c": "d"}},
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_valid_user_configuration_enum(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_choice": {"title": "Decisions", "type": "enum", "required": True, "default": 1, "choices": [1, 2, 3]}
                    }
                },
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_choice": {
                    "title": "Decisions",
                    "type": "enum",
                    "required": True,
                    "default": 1,
                    "choices": [1, 2, 3]
                }
            }
        })

    def test_create_valid_user_configuration_list(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": [1, 2, 3, 4]},
                        "boolean_list": {"title": "It's a list of booleans", "type": "list", "required": False, "subtype": "bool", "default": [True, False, False]}
                    }
                },
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_list": {
                    "title": "Demo List",
                    "type": "list",
                    "subtype": "int",
                    "required": True,
                    "default": [1, 2, 3, 4]
                },
                "boolean_list": {
                    "title": "It's a list of booleans",
                    "type": "list",
                    "subtype": "bool",
                    "required": False,
                    "default": [True, False, False]
                }
            }
        })

    def test_create_valid_user_configuration_model(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_model": {"title": "Model for training", "type": "model", "required": True},
                        "other_model": {"title": "Model the second", "type": "model", "default": str(self.model.id)}
                    }
                },
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_model": {
                    "title": "Model for training",
                    "type": "model",
                    "required": True
                },
                "other_model": {
                    "title": "Model the second",
                    "type": "model",
                    "default": str(self.model.id)
                }
            }
        })

    def test_create_invalid_user_configuration_list_requires_subtype(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "default": [1, 2, 3, 4]},
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "subtype": ['The "subtype" field must be set for "list" type properties.']
                    }
                }]
            }
        })

    def test_create_invalid_user_configuration_list_wrong_default(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": 12},
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "default": ["This is not a valid value for a field of type list."]
                    }
                }]
            }
        })

    def test_create_invalid_user_configuration_list_wrong_subtype(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "dict", "default": [1, 2, 3, 4]},
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "subtype": ["Subtype can only be int, float, bool or string."]
                    }
                }]
            }
        })

    def test_create_invalid_user_configuration_list_wrong_default_subtype(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": [1, 2, "three", 4]},
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "default": ["All items in the default value must be of type int."]
                    }
                }]
            }
        })

    def test_create_invalid_user_configuration_not_list_choices(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_choice": {"title": "Decisions", "type": "enum", "required": True, "default": 1, "choices": "eeee"}
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_choice": {
                        "choices": ['Expected a list of items but got type "str".']
                    }
                }]
            }
        })

    def test_create_invalid_user_configuration_not_dict(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": "non"
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": {"user_configuration": [['Expected a dictionary of items but got type "str".']]}})

    def test_create_invalid_user_configuration_item_not_dict(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "secrets": ["aaaaaa"]
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": {
            "user_configuration": [{
                "secrets":
                {"__all__": ["User configuration field definitions should be of type dict, not list."]}
            }]
        }})

    def test_create_invalid_user_configuration_wrong_field_type(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "something": {
                            "title": "some thing",
                            "type": "uh oh",
                            "required": 2
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "something": {
                        "required": ["Must be a valid boolean."],
                        "type": ["Value is not of type UserConfigurationFieldType"]
                    }
                }]
            }
        })

    def test_create_invalid_user_configuration_wrong_default_type(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "one_float": {
                            "title": "a float",
                            "type": "float",
                            "default": "bonjour",
                            "required": True
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "one_float": {
                        "default": ["This is not a valid value for a field of type float."]
                    }
                }]
            }
        })

    def test_create_invalid_user_configuration_choices_no_enum(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "something": {
                            "title": "some thing",
                            "type": "int",
                            "required": False,
                            "choices": [1, 2, 3]
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "something": {
                        "choices": ['The "choices" field can only be set for an "enum" type property.']
                    }
                }]
            }
        })

    def test_create_invalid_user_configuration_missing_key(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {"type": "int", "required": True, "default": 1}
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {
                "configuration": {
                    "user_configuration": [{
                        "demo_integer": {
                            "title": ["This field is required."]
                        }
                    }]
                }
            }
        )

    def test_create_invalid_user_configuration_invalid_key(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {
                            "title": "an integer",
                            "type": "int",
                            "required": True,
                            "default": 1,
                            "some_key": "oh no",
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [
                    {
                        "demo_integer": {
                            "some_key": ["Configurable properties can only be defined using the following keys: title, type, required, default, subtype, choices."]
                        }
                    }
                ]
            }
        })

    def test_create_invalid_user_configuration_default_value(self):
        cases = [
            ({"type": "int", "default": False}, "int"),
            ({"type": "int", "default": True}, "int"),
            ({"type": "float", "default": False}, "float"),
            ({"type": "float", "default": True}, "float"),
            ({"type": "bool", "default": 0}, "bool"),
            ({"type": "bool", "default": 1}, "bool"),
            ({"type": "string", "default": 1}, "string"),
            ({"type": "dict", "default": ["a", "b"]}, "dict"),
            ({"type": "model", "default": "gigi hadid"}, "model"),
            ({"type": "model", "default": False}, "model")
        ]
        for params, expected in cases:
            with self.subTest(**params):
                response = self.client.post(
                    reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
                    data={
                        "configuration": {
                            "user_configuration": {
                                "param": {"title": "param", **params}
                            }
                        },
                    },
                    HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), {"configuration": {"user_configuration": [{"param": {"default": [f"This is not a valid value for a field of type {expected}."]}}]}})

    def test_create_user_configuration_model_default_doesnt_exist(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "param": {"title": "Model to train", "type": "model", "default": "12341234-1234-1234-1234-123412341234"}
                    }
                },
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": {"user_configuration": [{"param": {"default": [
            "Model 12341234-1234-1234-1234-123412341234 not found."
        ]}}]}})
