from datetime import datetime, timezone
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.process.models import Worker, WorkerType
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model
from arkindex.users.models import Role


class TestWorkerList(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.worker_type_dla = WorkerType.objects.get(slug="dla")
        cls.worker_reco = Worker.objects.get(slug="reco")
        cls.worker_dla = Worker.objects.get(slug="dla")
        cls.worker_gpu = Worker.objects.get(slug="worker-gpu")
        cls.worker_generic = Worker.objects.get(slug="generic")
        cls.worker_file_import = Worker.objects.get(slug="file_import")
        cls.worker_custom = Worker.objects.get(slug="custom")
        cls.init_worker = Worker.objects.get(slug="initialisation")

        cls.model = Model.objects.create(
            name="Generic model",
            public=False,
        )

    def test_requires_login(self):
        response = self.client.get(reverse("api:workers-list"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 7,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_custom.id),
                    "repository_url": None,
                    "name": "Custom worker",
                    "description": "",
                    "slug": "custom",
                    "type": "custom",
                    "archived": False,
                },
                {
                    "id": str(self.worker_dla.id),
                    "repository_url": None,
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                },
                {
                    "id": str(self.init_worker.id),
                    "repository_url": None,
                    "name": "Elements Initialisation Worker",
                    "description": "",
                    "slug": "initialisation",
                    "type": "init",
                    "archived": False,
                },
                {
                    "id": str(self.worker_file_import.id),
                    "repository_url": None,
                    "name": "File import",
                    "description": "",
                    "slug": "file_import",
                    "type": "import",
                    "archived": False,
                },
                {
                    "id": str(self.worker_generic.id),
                    "repository_url": None,
                    "name": "Generic worker with a Model",
                    "description": "",
                    "slug": "generic",
                    "type": "recognizer",
                    "archived": False,
                },
                {
                    "id": str(self.worker_reco.id),
                    "repository_url": None,
                    "name": "Recognizer",
                    "description": "",
                    "slug": "reco",
                    "type": "recognizer",
                    "archived": False,
                },
                {
                    "id": str(self.worker_gpu.id),
                    "repository_url": None,
                    "name": "Worker requiring a GPU",
                    "description": "",
                    "slug": "worker-gpu",
                    "type": "worker",
                    "archived": False,
                },
            ]
        })

    def test_compatible_model(self):
        """
        User is able to filter workers by a compatible model.
        """
        self.model.memberships.create(user=self.user, level=Role.Guest.value)
        self.worker_generic.models.set([self.model])
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:workers-list"),
                {"compatible_model": str(self.model.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_generic.id),
                    "repository_url": None,
                    "name": "Generic worker with a Model",
                    "description": "",
                    "slug": "generic",
                    "type": "recognizer",
                    "archived": False,
                },
            ]
        })

    def test_compatible_model_public(self):
        public_model = Model.objects.create(name="Public", public=True)
        self.worker_reco.models.set([public_model])
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:workers-list"),
                {"compatible_model": str(public_model.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [{
                "id": str(self.worker_reco.id),
                "repository_url": None,
                "name": "Recognizer",
                "description": "",
                "slug": "reco",
                "type": "recognizer",
                "archived": False,
            }],
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_compatible_model_private(self, filter_rights_mock):
        private_model = Model.objects.create(name="Private")
        self.worker_reco.models.set([private_model])
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.all(),
            Model.objects.none(),
        ]

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:workers-list"),
                {"compatible_model": str(private_model.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "compatible_model": ["This model does not exist or you don't have access to it"]
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Model, Role.Guest.value),
        ])

    def test_filter_invalid_query_params(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:workers-list"),
                {"compatible_model": "A"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "compatible_model": ["Invalid UUID"],
        })

    def test_filter_type_slug(self):
        """
        User is able to filter workers by worker type slug
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:workers-list"), {"type": "dla"})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_url": None,
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                }
            ]
        })

    def test_filter_archived(self):
        self.client.force_login(self.user)
        self.worker_dla.archived = datetime.now(timezone.utc)
        self.worker_dla.save()

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"archived": True})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_url": None,
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": True,
                }
            ]
        })

    def test_filter_unarchived(self):
        self.client.force_login(self.user)
        Worker.objects.all().update(archived=datetime.now(timezone.utc))
        self.worker_dla.archived = None
        self.worker_dla.save()

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"archived": False})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_url": None,
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                }
            ]
        })

    def test_filter_invalid_type_slug(self):
        """
        Raise when trying to filter workers with an invalid type slug
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:workers-list"), {"type": "invalid-slug"})
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"type": ["No registered worker type with that slug."]})

    def test_filter_type_id(self):
        """
        User is able to filter workers by worker type id
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"type": self.worker_type_dla.id})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_url": None,
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                }
            ]
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_requires_contributor(self, filter_rights_mock):
        """
        User is not able to list workers with a guest access, except public workers
        """
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.none(),
        ]

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 2,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "archived": False,
                    "description": "",
                    "id": "bf56381e-a669-4e61-8d19-fc0e23171463",
                    "name": "Elements Initialisation Worker",
                    "repository_url": None,
                    "slug": "initialisation",
                    "type": "init"
                },
                {
                    "archived": False,
                    "description": "",
                    "id": "f0f5787d-3e41-4d16-aa56-b84f64103c12",
                    "name": "File import",
                    "repository_url": None,
                    "slug": "file_import",
                    "type": "import"
                }
            ]
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
        ])

    def test_name_filter(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"name": "layout"})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_url": None,
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                }
            ]
        })
